#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>


typedef struct _element {
    int data;
    struct _element *next;
    struct _element *prev;
} element;

typedef struct _dll{
    element* pos;
    element* tete;
}dll_t;


// crée la liste doublement chaînée
dll_t dll_create(){
    element *pos = NULL;
    element *tete = NULL;
    dll_t liste = {pos, tete};
    return liste;
}

// la liste est-elle vide?
bool dll_is_empty(dll_t list){
    return list.tete == NULL;
}

// retourne la valeur à la position actuelle dans la liste
int dll_value(dll_t list){
    assert(!dll_is_empty(list));
    if (list.pos == NULL){
        return 0;
    }
    return list.pos->data;
    
}


// Est-ce que pos est le 1er élément?
bool dll_is_head(dll_t list){
    assert(!dll_is_empty(list));
    return list.pos == list.tete;
}
// Est-ce que pos est le dernier élément?
bool dll_is_tail(dll_t list){
    assert(!dll_is_empty(list));
    if (list.pos != NULL){
        return list.pos->next == NULL;
    }
    return false;
    
}
// data est-elle dans la liste?
bool dll_is_present(dll_t list, int data){
    assert(!dll_is_empty(list));
    while (dll_value(list) != data){
        if(dll_is_tail(list)){
            return false;
        }
        list.pos = list.pos->next;
    }
    return true;
    
}

// affiche la liste
void dll_print(dll_t list){
    assert(!dll_is_empty(list));
    list.pos = list.tete;
    while (!dll_is_tail(list)){
        printf("%d ", dll_value(list));
        list.pos = list.pos->next;
    }
    printf("%d\n", dll_value(list));
}

// déplace la val de pos au début de la liste
dll_t dll_switch_to_head(dll_t list){
    assert(!dll_is_empty(list));
    if(dll_is_head(list)){
        return list;
    }
    if(dll_is_tail(list)){ // a finir
        list.pos->prev->next = NULL;
        
    }
    else{
        list.pos->prev->next = list.pos->next;
        list.pos->next->prev = list.pos->prev;
        
    }
    
    list.pos->next = list.tete;
    list.pos->prev = NULL;
    list.tete->prev = list.pos;
    list.tete = list.pos;
    return list;
}

dll_t dll_move_to_head(dll_t list){
    list.pos = list.tete;
    return list;
}

// déplace pos à la position suivante dans la liste
dll_t dll_next(dll_t list){
    if(!dll_is_tail(list)){
        list.pos = list.pos->next;
    }
    return list;
}
// déplace pos à la position précédente dans la liste
dll_t dll_prev(dll_t list){
    if(!dll_is_head(list)){
        list.pos = list.pos->prev;
    }
    return list;
}

// insertion de data dans l'élément après pos
dll_t dll_insert_after(dll_t list, int data){
    assert(!dll_is_empty(list));
    element *el = malloc(sizeof(element));
    assert(el != NULL);
    el->data = data;
    el->prev = list.pos;
    if (dll_is_tail(list)){
        el->next = NULL;
    }
    else {
        el->next = list.pos->next;
        el->next->prev = el;
    }
    list.pos->next = el;
    return list; 
}
// insertion de data en tête de liste
dll_t dll_push(dll_t list, int data){
    element *el = malloc(sizeof(element));
    assert(el != NULL);
    el->data = data;
    el->prev = NULL;
    if(!dll_is_empty(list)){
        el->next = list.tete;
        list.tete->prev = el;
    }
    else{
        el->next = NULL;
    }
    list.tete = el;
    return list;
}

// extraction de la valeur se trouvant dans l'élément pos
// l'élément pos est libéré
int dll_extract(dll_t *list){
    int val = dll_value(*list);
    if(dll_is_head(*list)){
        list->pos->next->prev = NULL;
        list->tete = list->pos->next;
    }
    else{
        list->pos->next->prev = list->pos->prev;
    }
    if(dll_is_tail(*list)){
        list->pos->prev->next = NULL;
    }
    else{
        
        list->pos->prev->next = list->pos->next;
    }
    element* tmp = list->pos->next;
    free(list->pos);
    list->pos = tmp;
    

    return val;
}
// extrait la donnée en tête de liste
int dll_pop(dll_t *list){
    assert(!dll_is_empty(*list));
    int val = list->tete->data;
    element* tmp;
    if (list->tete->next != NULL){
        list->tete->next->prev = NULL;
        tmp = list->tete->next;
        free(list->tete);
        list->tete = tmp;
    }
    else{
        free(list->tete);
        list->tete = NULL;
    }
    return val;
    
}
// vide la liste
void dll_destroy(dll_t *list){
    
    while(!dll_is_empty(*list)){
        dll_pop(list);
    }
}


int main(){
    dll_t list = dll_create();
    printf("vide : %d\n", dll_is_empty(list));
    list = dll_push(list,2);
    list = dll_push(list,7);
    list = dll_push(list,3);
    printf("is tail : %d\n",dll_is_tail(list));
    list = dll_move_to_head(list);
    printf("is head : %d\n",dll_is_head(list));
    printf("vide : %d\n", dll_is_empty(list));
    printf("value : %d\n", dll_value(list));
    printf("2 is present : %d\n",dll_is_present(list, 2));
    dll_print(list);
    list = dll_next(list);
    list = dll_insert_after(list, 4);
    dll_print(list);
    int val = dll_value(list);
    printf("the pos data is %d\n",val);
    val  = dll_extract(&list);
    printf("%d extracted from dll\n",val);
    dll_print(list);
    val = dll_pop(&list);
    printf("%d popped from dll\n",val);
    dll_print(list);
    dll_destroy(&list);
    printf("destroying the list, empty :%d\n",dll_is_empty(list));
    return 0;
}